"""API for the AnnotationTask
"""

class AnnotationTask():
    """Scalabel Annotation Task.
    """

    def __init__(self, annotation_name: str):
        """Instantiate new annotation tasks.
        """
        self.annotation_name = annotation_name

    def summarize(self) -> str:
        """Summarizing the object.
        """
        return self.annotation_name



"""
Define the logger format
"""

import logging


FORMAT = "[%(asctime)-15s %(filename)s:%(lineno)d %(funcName)s] %(message)s"
logging.basicConfig(format=FORMAT)
logger = logging.getLogger(__name__)  # pylint: disable=invalid-name
logger.setLevel(logging.DEBUG)



"""Datu API.
"""
from typing import Any, Dict, Optional
import os

import requests

from .common.enums import RequestType, RouteType


class Client():
    """API Client: shape the request, submit the request to the server and
    process the response.
    """

    def __init__(self, address: str = 'http://localhost', port: int = 5000):
        """Initializing the DatuClient.
        """
        self.address = address
        self.port = port
        self.server_address = '%s:%d' % (self.address, self.port)

    def send_request(self,
                     request_type: RequestType,
                     route: RouteType,
                     payload: Optional[Dict[str, Any]],
                     files: Optional[Dict[str, Any]] = None
                     ) -> Optional[Dict[str, Any]]:
        """Sending post request to the API server.

        Args:
            request_type: Specify whether it's a POST or GET request.
            route: The name of the route to be handled by the server.
            payload: The JSON payload to be sent off the API server.
            files: Special type for sending file object to API server.
        """
        address = os.path.join(self.server_address, route)
        if request_type == RequestType.POST:
            response = requests.post(address, json=payload, files=files)
        elif request_type == RequestType.GET:
            response = requests.get(address, json=payload)
        else:
            raise ValueError('request_type not supported')

        json_response: Dict[str, Any] = response.json()
        return json_response



"""Model definition for component of the Datu API.
"""
from __future__ import annotations

from typing import Any, Dict, Optional, List, Union
import json
import os
import typing

import jsonpickle
import tabulate

from .annotation_task import AnnotationTask
from .common.enums import (
    TaskType, SamplingStrategy, RequestType, RouteType)
from .common.logging import logger
from .trainer import Trainer
from .base_api import BaseAPI


JSONTYPE = Dict[str, Union[str, List[TaskType]]]


class Sample():
    """A sample of the data.
    """

    def __init__(self, name: str):
        self.name = name

    def __repr__(self) -> str:
        return "<Sample name='%s'" % self.name


class Project(BaseAPI):
    """A Datu project.
    """

    project_name: str
    project_id: str
    task_list: List[TaskType]

    def populate_from_json(self, json_data: JSONTYPE) -> None:
        """Populate the instance from the json.
        """
        required_fields = set(['project_name', 'project_id', 'task_list'])
        missing_keys = required_fields - set(json_data.keys())

        if missing_keys:
            logger.error('Required fields are missing: %s', str(missing_keys))
            return

        assert isinstance(json_data['project_name'], str)
        self.project_name = json_data['project_name']

        assert isinstance(json_data['project_id'], str)
        self.project_id = json_data['project_id']

        assert isinstance(json_data['task_list'], list)
        self.task_list = json_data['task_list']

    @classmethod
    def create_project(cls,
                       project_name: str,
                       task_list: List[TaskType],
                       categories: List[str]) -> 'Project':
        """Create a new project.
        Args:
            project_name: Name of the project.
            task_list: List of tasks to train for.
            categories: The list of categories to train for.

        Returns:
            A Project object.
        """
        payload = {
            'project_name': project_name,
            'task_list': task_list,
            'category_list': categories,
            'username': 'root',
        }

        json_data = cls.client.send_request(
            request_type=RequestType.POST,
            route=RouteType.CREATE_PROJECT,
            payload=payload)

        assert json_data is not None

        project_instance = cls()
        project_instance.populate_from_json(json_data)

        return project_instance

    @classmethod
    def from_id(cls, project_id: str) -> 'Optional[Project]':
        """Get a Project instance from an existing id.

        Args:
            project_id:
        """
        payload = {'project_id': project_id}
        json_data = cls.client.send_request(
            request_type=RequestType.GET,
            route=RouteType.PROJECT_LOOKUP,
            payload=payload)

        if not json_data:
            return None

        project_instance = cls()
        project_instance.populate_from_json(json_data)

        return project_instance

    @classmethod
    def list_projects(cls) -> None:
        """List the project created.
        """
        json_data = cls.client.send_request(
            request_type=RequestType.GET,
            route=RouteType.LIST_PROJECTS,
            payload=None,
        )

        assert json_data is not None

        project_list: List[Dict[str, Any]] = json_data['project_list']

        headers = ['project_id', 'project_name', 'tasks', 'status']
        rows = [[
            proj['project_id'],
            proj['project_name'],
            proj['task_list'],
            proj['status'],
        ] for proj in project_list]
        print(tabulate.tabulate(rows, headers=headers))

    def sample(self,
               sampling_strategy: SamplingStrategy,
               num_sample: int) -> typing.List[Sample]:
        """Sample the data from the existing data.

        Arguments:
            sampling_strategy: Specify the sampling strategy.
        """
        payload = {
            'project_id': self.project_id,
            'sampling_strategy': sampling_strategy,
            'num_sample': num_sample
        }

        json_data = self.client.send_request(
            request_type=RequestType.GET,
            route=RouteType.SAMPLING,
            payload=payload
        )

        assert json_data is not None

        sample_list = [
            Sample(file_name)
            for file_name in json_data['sample_list']
        ]

        return sample_list

    def upload_data(self, datapath: str) -> Optional[List[str]]:
        """Upload the data to the server.

        Arguments:
            dataset_name {str} -- Designated name for this dataset, it
                could be an s3 data path.
            datapath {str} -- Path to the tar file to upload.
        """
        if not os.path.exists(datapath):
            logger.error('%s is not available', datapath)
            return None

        size_in_mb = float(os.path.getsize(datapath)) / 1e6

        if size_in_mb > 100:
            logger.error("File size > 100MB, use upload_s3 instead.")
            return None

        with open(datapath, 'rb') as file_handle:
            files = {
                'file': file_handle,
                'data': json.dumps({
                    'project_id': self.project_id,
                    'username': 'root',
                })
            }

            json_data = self.client.send_request(
                request_type=RequestType.POST,
                route=RouteType.UPLOAD,
                payload=None,
                files=files)

            assert json_data is not None

            uploaded_file_list: List[str] = json_data['uploaded_files']

            logger.info(
                'Successfully uploaded the file, found %d image files.',
                len(uploaded_file_list))

            return uploaded_file_list

    def start_annotation(
            self,
            dataset_name: str,
            sample_list: typing.List[Sample],
            task_list: Union[List[TaskType], None] = None) -> AnnotationTask:
        """Start the annotation process.

        Args:
            dataset_name {str}: Name of the dataset.
            sample_list: A list of sample to be annotated by Scalabel.
            task_list: The list of tasks to annotate, if None, defaulted to the
                list from project creation.
        """
        payload = {
            "sample_list": jsonpickle.encode(sample_list),
            'dataset_name': dataset_name,
            "task_list": task_list,
            "project_id": self.project_id,
            "username": "root",
        }

        json_data = self.client.send_request(
            request_type=RequestType.POST,
            route=RouteType.START_ANNOTATION,
            payload=payload,
        )

        assert json_data is not None
        logger.info('New Scalabel task is initialized.')

        # TODO(phuc): Make a real annotation task. This makes a dummy
        # AnnotationTask for now.
        return AnnotationTask('abc')

    def create_trainer(self,
                       dataset_name: str,
                       train_val_split: float = 0.8,
                       min_annotation_per_class: int = 50) -> Trainer:
        """Start the training process using the annotated files.

        Args:
            dataset_name: Name of dataset.
            train_val_split: Split ratio between training and validation.
            min_annotation_per_class: Minimum number of annotations per class
                to start training.

        Returns:
            A Trainer object.
        """
        payload = {
            'username': 'root',
            'project_id': self.project_id,
            'dataset_name': dataset_name,
            'train_val_split': train_val_split,
            'min_annotation_per_class': min_annotation_per_class,
        }

        json_data = self.client.send_request(
            request_type=RequestType.POST,
            route=RouteType.CREATE_TRAINER,
            payload=payload
        )

        assert json_data is not None

        trainer = Trainer(client=self.client)
        trainer.populate_from_json(json_data)

        return trainer

    def list_trainers(self) -> bool:
        """Print the existing trainers for this project.
        """
        payload = {
            'project_id': self.project_id
        }

        json_data = self.client.send_request(
            request_type=RequestType.GET,
            route=RouteType.LIST_TRAINER,
            payload=payload
        )

        assert json_data is not None

        trainer_list = json_data['trainer_list']

        headers = ['trainer_id', 'model', 'dataset', 'tensorboard', 'status']
        rows = [
            [
                trainer_dict['trainer_id'],
                trainer_dict['model'],
                trainer_dict['dataset'],
                trainer_dict['tensorboard'],
                trainer_dict['status'],
            ]
            for trainer_dict in trainer_list
        ]
        print(tabulate.tabulate(rows, headers=headers))
        return True

    def __repr__(self) -> str:
        return '<Project %s project_id=%s> %s' % (
            self.project_name, self.project_id, str(self.task_list))



"""Class definition for the client-side Trainer object.
"""
from typing import Dict, Union
from .base_api import BaseAPI
from .common.logging import logger
from .common.enums import RequestType, RouteType


JSONTYPE = Dict[str, Union[str, str]]


class Trainer(BaseAPI):
    """Class definition for the trainer.
    """

    trainer_id: str
    project_id: str

    def populate_from_json(self, json_data: JSONTYPE) -> None:
        """Populate from json.
        """
        required_fields = set(['project_id', 'trainer_id'])
        missing_keys = required_fields - set(json_data.keys())

        if missing_keys:
            logger.error('Required fields are missing: %s', str(missing_keys))
            return

        assert isinstance(json_data['project_id'], str)
        self.project_id = json_data['project_id']

        assert isinstance(json_data['trainer_id'], str)
        self.trainer_id = json_data['trainer_id']

    def start_training(self) -> None:
        """Kickstart the training for this trainer.

        Args:
            force {bool} -- Whether to force the training even when the trainer
                is not ready.
        """
        payload = {
            'username': 'root',
            'trainer_id': self.trainer_id,
            'project_id': self.project_id,
        }
        json_data = self.client.send_request(
            request_type=RequestType.POST,
            route=RouteType.START_TRAINING,
            payload=payload
        )

        assert json_data is not None

        if json_data['status'] == 'success':
            logger.error('Error: %s', json_data['message'])



"""Model implementation.
"""
from __future__ import annotations
from typing import Dict, Any, Optional
from .client import Client

class BaseAPI():
    """Class definition for the model class.
    """

    client = Client()

    def __init__(self,
                 attrs: Optional[Dict[str, Any]] = None,
                 client: Optional[Client] = None):
        """Initiate the model with the client and the attributes.
        """
        if client:
            self.client = client

        self.attrs = attrs

    def summarize(self) -> str:
        """Summarize this object.
        """
        return str(self.attrs)



"""Define the Enumeration for the datu_api.
"""

from enum import Enum


# TODO(phuc): move to a different files.
class RequestType(Enum):
    """Different request types.
    """
    GET = 1
    POST = 2


class RouteType(str, Enum):
    """Different route names handled by the backend.
    """
    CREATE_PROJECT = 'create_project'
    PROJECT_LOOKUP = 'project_lookup'
    LIST_PROJECTS = 'list_projects'
    SAMPLING = 'sampling'
    UPLOAD = 'upload'
    START_ANNOTATION = 'start_annotation'
    CREATE_TRAINER = 'create_trainer'
    LIST_TRAINER = 'list_trainer'
    START_TRAINING = 'start_training'


class SamplingStrategy(str, Enum):
    """Enumeration for task type"""
    RANDOM = 'random'
    ACTIVE = 'active'


class DataType(str, Enum):
    """Type of input."""
    IMAGE = 'image'
    VIDEO = 'video'


class TaskType(str, Enum):
    """Enumeration for task type"""
    DETECTION_2D = 'detection2d'
    DETECTION_3D = 'detection3d'
    SEGMENTATION_SEMANTIC = 'segmentation_semantic'
    SEGMENTATION_INSTANCE = 'segmentation_instance'
    TRACKING_2D = 'tracking2d'
    TRACKING_3D = 'tracking3d'

class ModelType(str, Enum):
    """Enumeration for model type"""
    DETECTRON2 = "detectron2"
    CUSTOM = "custom"
